from __future__ import annotations

import uuid
from typing import Any


def first_not_null(*items: Any) -> Any:
    for item in items:
        if item is not None:
            return item
    return None


def generate_consumer_tag(*, prefix: str = "", channel_number: int = 1) -> str:
    return f"{prefix}ctag{channel_number}.{uuid.uuid4().hex}"



from __future__ import annotations

import asyncio
import email.utils
import importlib
import json
import operator
import re
import threading
from typing import Any

from ._state import _STATE


def _get_running_tasks() -> list[dict[str, Any]]:
    return [t.serialize() for t in _STATE.running_tasks.values()]


def _collect_running_tasks_statistics() -> dict[str, Any]:
    running_tasks = _get_running_tasks()

    def s(t: asyncio.Task[Any]) -> tuple[int, str]:
        n = t.get_name()
        match = re.fullmatch(r"Task-(\d+)", n)
        if match:
            return int(match[1]), n
        return 0, n

    asyncio_tasks = sorted(asyncio.all_tasks(), key=s)

    celery_sleeping_tasks = [t for t in running_tasks if t["state"] == "SLEEPING"]
    celery_semaphore_tasks = [t for t in running_tasks if t["state"] == "SEMAPHORE"]
    celery_gc_tasks = [t for t in running_tasks if t["state"] == "GC"]
    celery_running_tasks = [t for t in running_tasks if t["state"] == "RUNNING"]
    celery_running_tasks.sort(key=operator.itemgetter("started"))

    for t in celery_running_tasks:
        del t["received"]
        del t["state"]
        del t["eta"]

    result = {
        "asyncio": [repr(t) for t in asyncio_tasks],
        "celery": celery_running_tasks,
        "stats": {
            "asyncio": len(asyncio_tasks),
            "sleeping": len(celery_sleeping_tasks),
            "semaphore": len(celery_semaphore_tasks),
            "gc": len(celery_gc_tasks),
            "running": len(celery_running_tasks),
        },
        "threads": [repr(t) for t in threading.enumerate()],
    }

    try:
        psutil = importlib.import_module("psutil")
    except ImportError:
        pass
    else:
        this = psutil.Process()
        result["processes"] = [repr(p) for p in this.children(recursive=True)]
        result["sockets"] = [repr(c) for c in this.connections(kind="inet")]

    return result


async def inspection_http_handler(
    reader: asyncio.StreamReader,  # noqa: ARG001
    writer: asyncio.StreamWriter,
) -> None:
    date: str = email.utils.formatdate(timeval=None, localtime=False, usegmt=True)

    content: bytes = (
        json.dumps(
            _collect_running_tasks_statistics(),
            ensure_ascii=False,
            indent=4,
        )
        + "\n"
    ).encode("utf8")

    lines: list[bytes] = [
        b"HTTP/1.1 200 OK\r\n",
        b"Server: aio-celery\r\n",
        b"Date: " + date.encode("latin1") + b"\r\n",
        b"Content-Type: application/json; charset=utf-8\r\n",
        b"Content-Length: " + str(len(content)).encode("latin1") + b"\r\n",
        b"\r\n",
        content,
    ]

    for line in lines:
        writer.write(line)
    writer.write_eof()

    await writer.drain()

    writer.close()



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any

from ._state import get_current_app
from .utils import first_not_null

if TYPE_CHECKING:
    import datetime

    from .app import Celery
    from .result import AsyncResult


class Signature(dict[str, Any]):
    def __init__(
        self,
        task: str,
        args: tuple[Any, ...] | None = None,
        kwargs: dict[str, Any] | None = None,
        *,
        options: dict[str, Any] | None = None,
        app: Celery | None = None,
        **ex: Any,
    ) -> None:
        self._app = app
        super().__init__(
            task=task,
            args=tuple(args or ()),
            kwargs=kwargs or {},
            options=dict(options or {}, **ex),
            subtask_type=None,
            immutable=False,
        )

    @staticmethod
    def from_dict(d: dict[str, Any], *, app: Celery | None = None) -> Signature:
        return Signature(
            d["task"],
            args=d["args"],
            kwargs=d["kwargs"],
            options=d["options"],
            app=app,
        )

    def set(self, **options: Any) -> Signature:
        self["options"].update(options)
        return self

    async def apply_async(  # noqa: PLR0913
        self,
        args: tuple[Any, ...] | None = None,
        kwargs: dict[str, Any] | None = None,
        *,
        task_id: str | None = None,
        countdown: float | None = None,
        priority: int | None = None,
        queue: str | None = None,
        expiration: datetime.datetime | datetime.timedelta | float | None = None,
    ) -> AsyncResult:
        sig: Signature = self
        chain_: list[dict[str, Any]] | None = None
        if self["task"] == "celery.chain":
            tasks = self["kwargs"]["tasks"]
            sig = tasks[0]
            chain_ = list(reversed(tasks[1:])) or None

        options = self["options"]

        return await (self._app or get_current_app()).send_task(
            sig["task"],
            args=first_not_null(args, sig["args"]),
            kwargs=first_not_null(kwargs, sig["kwargs"]),
            countdown=first_not_null(countdown, options.get("countdown")),
            task_id=first_not_null(task_id, options.get("task_id")),
            priority=first_not_null(priority, options.get("priority")),
            queue=first_not_null(queue, options.get("queue")),
            chain=chain_,
            expiration=first_not_null(expiration, options.get("expiration")),
        )

    async def delay(self, *args: Any, **kwargs: Any) -> AsyncResult:
        return await self.apply_async(args=args, kwargs=kwargs)


def signature(
    varies: str | Signature | dict[str, Any],
    *args: Any,
    **kwargs: Any,
) -> Signature:
    app = kwargs.get("app")
    if isinstance(varies, dict):
        if isinstance(varies, Signature):
            return copy.deepcopy(varies)
        return Signature.from_dict(varies, app=app)
    return Signature(varies, *args, **kwargs)


def chain(*tasks: Signature, **kwargs: Any) -> Signature:
    return Signature("celery.chain", args=(), kwargs={"tasks": tasks}, **kwargs)



from __future__ import annotations

import argparse
import asyncio
import logging
import sys

from . import __version__
from .worker import run


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="aio_celery",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(
            "CLI for aio_celery\n\n"
            "Examples\n"
            "--------\n"
            "$ %(prog)s worker proj:app\n"
            "$ %(prog)s worker -Q high,low proj:app\n"
            "$ %(prog)s worker --concurrency=50000 proj:app\n"
        ),
    )
    parser.add_argument(
        "-V",
        "--version",
        dest="version",
        action="store_true",
        help="print current aio_celery version and exit",
    )
    subparsers = parser.add_subparsers(
        title="Available subcommands",
        metavar="",
        dest="subcommand",
    )
    worker = subparsers.add_parser(
        "worker",
        help="Run worker",
        description="Run worker",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    worker.add_argument(
        "app",
        help=(
            "Where to search for application instance. "
            "This string must be specified in 'module.module:variable' format."
        ),
    )
    worker.add_argument(
        "-c",
        "--concurrency",
        type=int,
        default=10_000,
        help="Maximum simultaneous async tasks",
    )
    worker.add_argument(
        "-Q",
        "--queues",
        help="Comma separated list of queues",
    )
    worker.add_argument(
        "-l",
        "--loglevel",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "FATAL"],
        default="INFO",
        help="Logging level",
    )
    worker.add_argument(
        "--no-configure-logging",
        action="store_false",
        dest="configure_logging",
        help="Use this parameter if your application configures custom logging.",
    )

    args = parser.parse_args()

    if args.version:
        print(__version__)  # noqa: T201
        return

    if args.subcommand is None:
        parser.print_help()
        return

    if args.concurrency < 1:
        print(  # noqa: T201
            "concurrency must be larger than 1",
            file=sys.stderr,
        )
        sys.exit(2)

    try:
        asyncio.run(run(args))
    except KeyboardInterrupt:
        logging.getLogger(__name__).warning("Worker process interrupted.")


if __name__ == "__main__":
    main()



from __future__ import annotations

from ._state import get_current_app as current_app
from .app import Celery, shared_task
from .canvas import chain, signature
from .task import Task

__version__ = "0.19.0"

__all__ = (
    "Celery",
    "Task",
    "chain",
    "current_app",
    "shared_task",
    "signature",
)



from __future__ import annotations

import asyncio
import asyncio.exceptions
import contextlib
import datetime
import functools
import importlib
import logging
import pathlib
import sys
import time
import urllib.parse
from typing import TYPE_CHECKING, Any, Awaitable, Iterator, Optional, cast

import aio_pika
import aiormq.exceptions
from aio_pika.abc import AbstractRobustChannel
from yarl import URL

from ._state import _STATE, RunningTask
from .app import Celery
from .context import CURRENT_ROOT_ID, CURRENT_TASK_ID
from .exceptions import MaxRetriesExceededError, Retry
from .inspection_http_server import inspection_http_handler
from .intermittent_gc import perform_gc_at_regular_intervals
from .request import Request
from .task import Task
from .utils import first_not_null, generate_consumer_tag

if TYPE_CHECKING:
    import argparse
    import types

    from aio_pika import IncomingMessage

    from .annotated_task import AnnotatedTask

sentry_sdk: types.ModuleType | None
try:
    sentry_sdk = importlib.import_module("sentry_sdk")
except ModuleNotFoundError:
    sentry_sdk = None

BANNER = """\
[config]
.> app:                 {app}
.> transport:           {conninfo}
.> results:             {results}
.> inspect http server: {inspect_http_server}
.> concurrency:         {concurrency}

[queues]
{queues}
"""

MAX_AMQP_PREFETCH_COUNT = 65535

logger = logging.getLogger(__name__)

_background_tasks = set()


def _iso_now() -> str:
    return datetime.datetime.now().astimezone().isoformat()


async def _sleep_if_necessary(task: Task) -> None:
    eta: datetime.datetime | None = task.request.eta
    if eta is None:
        return
    now = datetime.datetime.now().astimezone()
    if eta > now:
        delta: datetime.timedelta = eta - now
        logger.debug(
            "Task %s[%s] Sleeping for %s until task ETA %s ...",
            task.request.task,
            task.request.id,
            delta,
            eta,
        )
        await asyncio.sleep(delta.total_seconds())
        logger.debug(
            "Task %s[%s] Sleeping FINISHED",
            task.request.task,
            task.request.id,
        )


async def _handle_task_result(
    task: Task,
    annotated_task: AnnotatedTask,
    result: Any,
) -> None:
    if (
        task.app.conf.task_ignore_result
        or annotated_task.ignore_result
        or task.request.ignore_result
    ):
        pass
    else:
        await task.update_state(state="SUCCESS", meta=result, _finalize=True)
    next_message, next_routing_key = task.build_next_task_message(result=result)
    if next_message is not None:
        await task.app.broker.publish_message(
            next_message,
            routing_key=next_routing_key,
        )


async def _handle_task_retry(
    *,
    task: Task,
    annotated_task: AnnotatedTask,
    app: Celery,
    exc: Retry,
) -> None:
    logger.info(
        "Task %s[%s] retry: %s",
        task.name,
        task.request.id,
        exc,
    )
    max_retries = annotated_task.max_retries
    if max_retries is not None and task.request.retries > max_retries:
        msg = (
            f"Can't retry {task.name}[{task.request.id}] args:{task.request.args}"
            f" kwargs:{task.request.kwargs} max_retries:{max_retries}"
        )
        raise MaxRetriesExceededError(msg)
    await app.broker.publish_message(
        exc.message,
        routing_key=(task.request.message.routing_key or app.conf.task_default_queue),
    )


def _get_soft_time_limit(task: Task, annotated_task: AnnotatedTask) -> float | None:
    return cast(
        Optional[float],
        first_not_null(
            task.request.timelimit[0],
            annotated_task.soft_time_limit,
            task.app.conf.task_soft_time_limit,
        ),
    )


def _setup_sentry_context(task: Task, annotated_task: AnnotatedTask) -> None:
    if sentry_sdk is None:
        return
    sentry_sdk.set_tag("aio_celery_task_id", task.request.id)
    sentry_sdk.set_context(
        "aio-celery task",
        {
            "args": task.request.args,
            "kwargs": task.request.kwargs,
            "task_name": task.name,
            "soft_time_limit": _get_soft_time_limit(task, annotated_task),
        },
    )


async def _run_with_timeout(coro: Awaitable[Any], soft_time_limit: float | None) -> Any:
    if soft_time_limit is None:
        return await coro
    if sys.version_info >= (3, 11):
        # https://stackoverflow.com/a/47004339
        async with asyncio.timeout(soft_time_limit):
            return await coro
    return await asyncio.wait_for(coro, timeout=soft_time_limit)


def _log_unregistered_task(message: IncomingMessage) -> None:
    logger.exception(
        "Received unregistered task of type %r\n"
        "The message has been ignored and discarded.\n\n"
        "Did you remember to import the module containing this task?\n"
        "Or maybe you're using relative imports?\n\n"
        "Please see\n"
        "https://docs.celeryq.dev/en/latest/internals/protocol.html\n"
        "for more information.\n\n"
        "The full contents of the message body was:\n"
        "%r (%db)\n\n"
        "The full contents of the message headers:\n"
        "%r\n\n"
        "The delivery info for this task is:\n"
        "%r",
        str(message.headers["task"]),
        message.body,
        len(message.body),
        message.headers,
        {
            "consumer_tag": message.consumer_tag,
            "delivery_tag": message.delivery_tag,
            "redelivered": message.redelivered,
            "exchange": message.exchange,
            "routing_key": message.routing_key,
        },
    )


async def _execute_task(
    task: Task,
    annotated_task: AnnotatedTask,
    message: IncomingMessage,
) -> None:
    app = task.app
    args = (task, *task.request.args) if annotated_task.bind else task.request.args
    logger.debug(
        "Task %s[%s] started executing after acquiring semaphore",
        task.name,
        task.request.id,
    )
    timestamp_start = time.monotonic()
    try:
        try:
            result = await _run_with_timeout(
                annotated_task.fn(*args, **task.request.kwargs),
                _get_soft_time_limit(task, annotated_task),
            )
        except Retry:
            raise
        except Exception as exc:
            if isinstance(exc, annotated_task.autoretry_for) or (
                isinstance(exc, asyncio.TimeoutError)
                and app.conf.worker_retry_task_on_asyncio_timeout_error
            ):
                logger.warning(
                    "Task %s[%s] raised asyncio.TimeoutError, retrying...",
                    task.name,
                    task.request.id,
                )
                await task.retry()
            else:
                await message.reject()
            raise
    except Retry as exc:
        try:
            await _handle_task_retry(
                task=task,
                annotated_task=annotated_task,
                app=app,
                exc=exc,
            )
        except MaxRetriesExceededError:
            await message.reject()
            raise
    else:
        await _handle_task_result(
            task=task,
            annotated_task=annotated_task,
            result=result,
        )
        logger.info(
            "Task %s[%s] succeeded in %.6fs: %r",
            task.name,
            task.request.id,
            time.monotonic() - timestamp_start,
            result,
        )


async def on_message_received(
    message: IncomingMessage,
    *,
    app: Celery,
    semaphore: asyncio.Semaphore,
    gc_is_paused: asyncio.Event,
) -> None:
    with contextlib.ExitStack() as stack:
        if app.conf.enable_sentry_sdk and sentry_sdk is not None:
            stack.enter_context(sentry_sdk.Hub(sentry_sdk.Hub.current))

        task_id = str(message.headers["id"])
        task_name = str(message.headers["task"])

        CURRENT_ROOT_ID.set(cast(Optional[str], message.headers["root_id"]))
        CURRENT_TASK_ID.set(task_id)

        logger.info("Task %s[%s] received", task_name, task_id)

        try:
            async with message.process(ignore_processed=True, requeue=True):
                try:
                    annotated_task = app.get_annotated_task(task_name)
                except KeyError:
                    _log_unregistered_task(message)
                    await message.reject()
                    return

                task = Task(
                    app=app,
                    request=Request.from_message(message),
                    _default_retry_delay=annotated_task.default_retry_delay,
                )

                running_task = RunningTask(
                    asyncio_task=asyncio.current_task(),  # type: ignore[arg-type]
                    task_id=task_id,
                    task_name=task_name,
                    received=_iso_now(),
                    state="SLEEPING",
                    args=task.request.args,
                    kwargs=task.request.kwargs,
                    eta=(
                        task.request.eta.isoformat()
                        if isinstance(task.request.eta, datetime.datetime)
                        else None
                    ),
                    soft_time_limit=_get_soft_time_limit(
                        task,
                        annotated_task,
                    ),
                    retries=task.request.retries,
                )
                _STATE.running_tasks[task_id] = running_task
                await asyncio.sleep(0)

                if app.conf.enable_sentry_sdk and sentry_sdk is not None:
                    _setup_sentry_context(task, annotated_task)

                logger.debug(
                    "Task %s[%s] args=%r kwargs=%r",
                    task.name,
                    task.request.id,
                    task.request.args,
                    task.request.kwargs,
                )

                await _sleep_if_necessary(task)
                running_task.state = "SEMAPHORE"

                async with semaphore:
                    running_task.state = "GC"
                    await gc_is_paused.wait()
                    running_task.state = "RUNNING"
                    running_task.started = _iso_now()
                    await _execute_task(task, annotated_task, message)
        except aiormq.exceptions.ChannelInvalidStateError:
            pass
        except Exception:
            # This does not catch asyncio.exceptions.CancelledError since
            # the latter inherits directly from BaseException — not from Exception.
            # And that is okay for us. CancelledError is not meant to be caught.
            logger.exception(
                "Task %s[%s] raised unexpected:",
                task_name,
                task_id,
            )
        finally:
            _STATE.amount_of_tasks_completed_after_last_gc_run += 1
            with contextlib.suppress(KeyError):
                del _STATE.running_tasks[task_id]
            await asyncio.sleep(0)
            CURRENT_ROOT_ID.set(None)
            CURRENT_TASK_ID.set(None)


@contextlib.contextmanager
def _cwd_in_path() -> Iterator[None]:
    """Context adding the current working directory to sys.path."""
    cwd = str(pathlib.Path.cwd())
    if cwd in sys.path:
        yield
    else:
        sys.path.insert(0, cwd)
        try:
            yield
        finally:
            with contextlib.suppress(ValueError):
                sys.path.remove(cwd)


def _find_app_instance(locator: str) -> Celery:
    with _cwd_in_path():
        module_name, app_name = locator.rsplit(":", 1)
        module = importlib.import_module(module_name)
        app = getattr(module, app_name)
    if not isinstance(app, Celery):
        msg = "app instance must inherit from Celery class"
        raise TypeError(msg)
    return app


def _print_intro(
    concurrency: int,
    queue_names: list[str],
    task_names: list[str],
    app: Celery,
) -> None:
    q = ".> " + "\n.> ".join(queue_names)
    t = "  . " + "\n  . ".join(task_names)

    if app.conf.inspection_http_server_is_enabled:
        inspect_http_server = (
            "http://"
            f"{app.conf.inspection_http_server_host}:"
            f"{app.conf.inspection_http_server_port}/"
        )
    else:
        inspect_http_server = "disabled://"

    def _normalize(u: str | None) -> str:
        if not u:
            return "disabled://"
        parts = urllib.parse.urlparse(u)
        if not parts.password:
            return u
        return str(URL(u).with_password("******"))

    banner = BANNER.format(
        app=repr(app),
        conninfo=_normalize(app.conf.broker_url),
        results=_normalize(app.conf.result_backend),
        concurrency=f"{concurrency} (asyncio)",
        queues=q,
        inspect_http_server=inspect_http_server,
    ).splitlines()
    print(  # noqa: T201
        "╔═══════════════════════╗\n"
        "║ AsyncIO Celery Worker ║\n"
        "╚═══════════════════════╝\n",
    )
    print("\n".join(banner) + "\n")  # noqa: T201
    print(f"[tasks]\n{t}\n")  # noqa: T201


async def run(args: argparse.Namespace) -> None:
    if args.configure_logging:
        logging.basicConfig(
            level=logging.getLevelName(args.loglevel),
            format="[%(asctime)s][%(name)s][%(levelname)s] %(message)s",
        )
    logging.getLogger("aio_celery").setLevel(level=logging.getLevelName(args.loglevel))
    app = _find_app_instance(args.app)
    if not args.queues:
        queue_names = [app.conf.task_default_queue]
    else:
        queue_names = [q.strip() for q in args.queues.split(",")]

    _print_intro(
        args.concurrency,
        queue_names,
        app.list_registered_task_names(),
        app,
    )

    gc_is_paused = asyncio.Event()
    gc_is_paused.set()
    if app.conf.intermittent_gc_is_enabled:
        gc_task = asyncio.create_task(
            perform_gc_at_regular_intervals(
                max_tasks_between_gc=app.conf.intermittent_gc_max_tasks_between_gc,
                max_interval_between_gc=app.conf.intermittent_gc_max_interval_between_gc,
                gc_is_paused=gc_is_paused,
            ),
        )
        _background_tasks.add(gc_task)
        gc_task.add_done_callback(_background_tasks.discard)

    async with app.setup():
        server = await asyncio.start_server(
            inspection_http_handler,
            host=app.conf.inspection_http_server_host,
            port=app.conf.inspection_http_server_port,
            start_serving=False,
        )

        semaphore = asyncio.Semaphore(args.concurrency)
        prefetch_count = min(
            app.conf.worker_prefetch_multiplier * args.concurrency,
            MAX_AMQP_PREFETCH_COUNT,
        )
        connection = await aio_pika.connect_robust(app.conf.broker_url)
        async with server, connection, connection.channel() as channel:
            if app.conf.inspection_http_server_is_enabled:
                await server.start_serving()
            if prefetch_count > 0:
                await channel.set_qos(
                    prefetch_count=prefetch_count,
                    timeout=app.conf.broker_publish_timeout,
                )
            queues = [
                await app.broker.declare_queue(
                    queue_name=name,
                    channel=cast(AbstractRobustChannel, channel),
                )
                for name in queue_names
            ]
            for queue in queues:
                if app.conf.worker_consumer_tag_prefix:
                    consumer_tag = generate_consumer_tag(
                        prefix=app.conf.worker_consumer_tag_prefix,
                        channel_number=1,
                    )
                else:
                    consumer_tag = None
                await queue.consume(
                    functools.partial(
                        on_message_received,
                        app=app,
                        semaphore=semaphore,
                        gc_is_paused=gc_is_paused,
                    ),
                    consumer_tag=consumer_tag,
                    timeout=app.conf.broker_publish_timeout,
                )
            logger.info("Waiting for messages. To exit press CTRL+C")
            await asyncio.Future()



from __future__ import annotations

import copy
import dataclasses
import datetime
import json
import logging
import uuid
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from .amqp import create_task_message
from .exceptions import Retry
from .utils import first_not_null

if TYPE_CHECKING:
    from aio_pika import Message

    from .app import Celery
    from .request import Request

logger = logging.getLogger(__name__)


@dataclass
class Task:
    app: Celery
    request: Request
    _default_retry_delay: float = dataclasses.field(repr=False)

    @property
    def name(self) -> str:
        return self.request.task

    async def update_state(
        self,
        *,
        state: str,
        meta: dict[str, Any],
        _finalize: bool = False,
    ) -> None:
        result_backend = self.app.result_backend
        if result_backend is None:
            logger.debug("Result backend has not been enabled")
            return
        if _finalize:
            # do not update if result already in redis
            pass
        payload: dict[str, Any] = {
            "status": state,
            "result": meta,
            "traceback": None,
            "children": [],
            "date_done": None,
            "task_id": self.request.id,
        }
        if _finalize:
            payload["date_done"] = datetime.datetime.now(
                datetime.timezone.utc,
            ).isoformat()
        if self.request.group is not None:
            payload["group_id"] = self.request.group
        if self.request.parent_id is not None:
            payload["parent_id"] = self.request.parent_id
        await result_backend.set(
            f"celery-task-meta-{self.request.id}",
            json.dumps(payload).encode(),
            ex=self.app.conf.result_expires,
        )

    def build_next_task_message(self, result: Any) -> tuple[Message | None, str]:
        if not self.request.chain:
            return None, ""
        new_chain = copy.deepcopy(self.request.chain)
        first = new_chain.pop()
        opts = first.get("options", {})
        new_task_id: str = first_not_null(opts.get("task_id"), str(uuid.uuid4()))
        new_priority: int = first_not_null(
            opts.get("priority"),
            self.app.conf.task_default_priority,
        )
        routing_key: str = first_not_null(
            opts.get("queue"),
            self.app.conf.task_default_queue,
        )
        reply_to: str = first_not_null(opts.get("reply_to"), "")
        expiration = opts.get("expiration")
        return (
            create_task_message(
                task_id=new_task_id,
                task_name=first["task"],
                args=(result, *first["args"]),
                kwargs=first["kwargs"],
                priority=new_priority,
                parent_id=self.request.id,
                root_id=self.request.root_id,
                chain=new_chain,
                reply_to=reply_to,
                expiration=expiration,
            ),
            routing_key,
        )

    async def retry(self, *, countdown: float | None = None) -> None:
        delay = datetime.timedelta(
            seconds=first_not_null(
                countdown,
                self._default_retry_delay,
            ),
        )
        raise Retry(
            message=self.request.build_retry_message(countdown=delay),
            delay=delay,
        )



from __future__ import annotations

import dataclasses
import datetime
from dataclasses import dataclass


@dataclass
class DefaultConfig:
    # Task execution settings
    task_ignore_result: bool = False
    task_soft_time_limit: float | None = None

    # Task result backend settings
    result_backend: str | None = None
    result_backend_connection_pool_size: int = 50
    result_expires: datetime.timedelta = datetime.timedelta(days=1)

    # Message Routing
    task_default_priority: int | None = None
    task_default_queue: str = "celery"
    task_queue_max_priority: int | None = None

    # Broker Settings
    broker_publish_timeout: float | None = None
    broker_url: str = "amqp://guest:guest@localhost:5672//"

    # Worker
    worker_consumer_tag_prefix: str = ""
    worker_prefetch_multiplier: int = 4
    # Raison d'être:
    # The asyncio.timeout() context manager is what transforms
    # the asyncio.CancelledError into a TimeoutError,
    # which means the TimeoutError can only be caught outside of the context manager.
    # (https://docs.python.org/3/library/asyncio-task.html#asyncio.timeout)
    worker_retry_task_on_asyncio_timeout_error: bool = False

    # Sentry SDK
    enable_sentry_sdk: bool = False

    # Inspection HTTP server
    inspection_http_server_is_enabled: bool = False
    inspection_http_server_host: str = "localhost"
    inspection_http_server_port: int = 1430

    # Intermittent Garbage Collection Behavior
    intermittent_gc_is_enabled: bool = False
    intermittent_gc_max_tasks_between_gc: int = 100
    intermittent_gc_max_interval_between_gc: datetime.timedelta = datetime.timedelta(
        hours=1,
    )

    def update(self, **options: int | bool | str | datetime.timedelta | None) -> None:
        fields = {f.name for f in dataclasses.fields(self.__class__)}
        for k, v in options.items():
            if k not in fields:
                msg = f"Unknown configuration option: {k!r}"
                raise ValueError(msg)
            setattr(self, k, v)



from __future__ import annotations

import asyncio
import contextlib
import logging
import sys
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    cast,
)

import aio_pika
from aio_pika.abc import AbstractRobustChannel

from ._state import set_current_app
from .amqp import create_task_message
from .annotated_task import AnnotatedTask
from .backend import create_redis_connection_pool
from .broker import Broker
from .config import DefaultConfig
from .context import CURRENT_ROOT_ID, CURRENT_TASK_ID
from .result import AsyncResult as _AsyncResult
from .utils import first_not_null

if TYPE_CHECKING:
    import datetime

    import redis.asyncio

logger = logging.getLogger(__name__)


class Celery:
    def __init__(self, name: str | None = None) -> None:
        self.name = name
        self.conf = DefaultConfig()
        self._tasks_registry: dict[str, AnnotatedTask] = {}
        self._app_context: Any = None
        self._result_backend_connection_pool: (  # type: ignore[type-arg]
            redis.asyncio.BlockingConnectionPool | None
        ) = None
        self._broker: Broker | None = None
        self._setup_app_context: Callable[
            [],
            contextlib.AbstractAsyncContextManager[Any],
        ] = _setup_nothing

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name} at {hex(id(self))}>"

    def define_app_context(
        self,
        fn: Callable[[], contextlib.AbstractAsyncContextManager[Any]],
    ) -> None:
        self._setup_app_context = fn

    @property
    def broker(self) -> Broker:
        if self._broker is None:
            msg = "This app has not been configured yet."
            raise RuntimeError(msg)
        return self._broker

    @broker.setter
    def broker(self, b: Broker) -> None:
        self._broker = b

    @property
    def context(self) -> Any:
        return self._app_context

    @property
    def result_backend(self) -> redis.asyncio.Redis[bytes] | None:
        if self._result_backend_connection_pool is None:
            return None

        from redis.asyncio import Redis  # noqa: PLC0415

        return Redis(connection_pool=self._result_backend_connection_pool)

    @contextlib.asynccontextmanager
    async def setup(self) -> AsyncIterator[None]:
        connection = await aio_pika.connect_robust(self.conf.broker_url)
        async with connection, connection.channel() as publishing_channel:
            self.broker = Broker(
                broker_url=self.conf.broker_url,
                broker_publish_timeout=self.conf.broker_publish_timeout,
                task_queue_max_priority=self.conf.task_queue_max_priority,
                publishing_channel=cast(AbstractRobustChannel, publishing_channel),
            )
            set_current_app(self)
            try:
                if self.conf.result_backend is not None:
                    self._result_backend_connection_pool = create_redis_connection_pool(
                        url=self.conf.result_backend,
                        pool_size=self.conf.result_backend_connection_pool_size,
                    )
                async with self._setup_app_context() as context:
                    self._app_context = context
                    try:
                        yield
                    finally:
                        logger.warning("Shutting down application.")
            finally:
                set_current_app(None)
                if self._result_backend_connection_pool is not None:
                    await self._result_backend_connection_pool.disconnect()

    def task(  # noqa: PLR0913
        self,
        *args: Callable[..., Awaitable[Any]],
        bind: bool = False,
        name: str | None = None,
        ignore_result: bool | None = None,
        max_retries: int | None = 3,
        default_retry_delay: float = 180,
        autoretry_for: tuple[type[Exception], ...] = (),
        queue: str | None = None,
        priority: int | None = None,
        soft_time_limit: float | None = None,
    ) -> AnnotatedTask | Callable[[Callable[..., Awaitable[Any]]], AnnotatedTask]:
        """Create a task class out of any callable."""

        def decorator(fn: Callable[..., Awaitable[Any]]) -> AnnotatedTask:
            if name is None:
                task_name = _gen_task_name(fn.__name__, fn.__module__)
            else:
                task_name = name
            if not asyncio.iscoroutinefunction(fn):
                msg_ = f"Task {task_name!r} ({fn}) must be a coroutine"
                raise TypeError(msg_)
            annotated_task = AnnotatedTask(
                fn=fn,
                bind=bind,
                ignore_result=ignore_result,
                max_retries=max_retries,
                default_retry_delay=default_retry_delay,
                autoretry_for=autoretry_for,
                name=task_name,
                queue=queue,
                priority=priority,
                soft_time_limit=soft_time_limit,
                app=self,
            )
            self._tasks_registry[task_name] = annotated_task
            return annotated_task

        if not args:
            return decorator
        if len(args) == 1:
            func = args[0]
            if not callable(func):
                msg = "argument 1 to @task() must be a callable"
                raise TypeError(msg)
            return decorator(func)
        msg = "@task() takes exactly 1 argument"
        raise TypeError(msg)

    def _construct_extended_task_registry(self) -> dict[str, AnnotatedTask]:
        registry: dict[str, AnnotatedTask] = {}
        for name, task in _SHARED_APP._tasks_registry.items():  # noqa: SLF001
            task.app = self
            registry[name] = task
        registry.update(self._tasks_registry)
        return registry

    def get_annotated_task(self, task_name: str) -> AnnotatedTask:
        return self._construct_extended_task_registry()[task_name]

    def list_registered_task_names(self) -> list[str]:
        return sorted(self._construct_extended_task_registry())

    def AsyncResult(self, task_id: str) -> _AsyncResult:  # noqa: N802
        return _AsyncResult(task_id, app=self)

    async def send_task(  # noqa: PLR0913
        self,
        name: str,
        *,
        args: tuple[Any, ...] | None = None,
        kwargs: dict[str, Any] | None = None,
        countdown: float | None = None,
        task_id: str | None = None,
        priority: int | None = None,
        queue: str | None = None,
        chain: list[dict[str, Any]] | None = None,
        expiration: datetime.datetime | datetime.timedelta | float | None = None,
    ) -> _AsyncResult:
        if task_id is None:
            task_id = str(uuid.uuid4())
        await self.broker.publish_message(
            create_task_message(
                task_id=task_id,
                task_name=name,
                args=args,
                kwargs=kwargs,
                priority=first_not_null(priority, self.conf.task_default_priority),
                countdown=countdown,
                parent_id=CURRENT_TASK_ID.get(),
                root_id=CURRENT_ROOT_ID.get(),
                chain=chain,
                expiration=expiration,
            ),
            routing_key=first_not_null(queue, self.conf.task_default_queue),
        )
        return self.AsyncResult(task_id)


@contextlib.asynccontextmanager
async def _setup_nothing() -> AsyncIterator[None]:  # noqa: RUF029
    yield None


def _gen_task_name(name: str, module_name: str) -> str:
    """Generate task name from name/module pair."""
    module_name = module_name or "__main__"
    module = sys.modules[module_name]
    if module is not None:
        module_name = module.__name__
    return ".".join(p for p in (module_name, name) if p)


_SHARED_APP = Celery()


def shared_task(
    *args: Callable[..., Awaitable[Any]],
    **kwargs: Any,
) -> AnnotatedTask | Callable[[Callable[..., Awaitable[Any]]], AnnotatedTask]:
    return _SHARED_APP.task(*args, **kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import redis.asyncio


def create_redis_connection_pool(
    *,
    url: str,
    pool_size: int,
) -> redis.asyncio.BlockingConnectionPool:  # type: ignore[type-arg]
    import redis.backoff  # noqa: PLC0415
    from redis.asyncio import BlockingConnectionPool  # noqa: PLC0415
    from redis.asyncio.retry import Retry  # noqa: PLC0415

    default_base = getattr(redis.backoff, "DEFAULT_BASE", 0.008)
    default_cap = getattr(redis.backoff, "DEFAULT_CAP", 0.512)

    return BlockingConnectionPool.from_url(
        url=url,
        max_connections=pool_size,
        timeout=None,
        socket_timeout=5,
        retry_on_timeout=True,
        retry=Retry(
            backoff=redis.backoff.EqualJitterBackoff(
                cap=default_cap,
                base=default_base,
            ),
            retries=5,
        ),
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Awaitable, Callable

from .canvas import Signature
from .utils import first_not_null

if TYPE_CHECKING:
    from .app import Celery
    from .result import AsyncResult


@dataclass
class AnnotatedTask:
    fn: Callable[..., Awaitable[Any]]
    bind: bool
    ignore_result: bool | None
    max_retries: int | None
    default_retry_delay: float
    autoretry_for: tuple[type[Exception], ...]
    app: Celery
    name: str
    queue: str | None
    priority: int | None
    soft_time_limit: float | None

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self.bind:
            return self.fn(self, *args, **kwargs)
        return self.fn(*args, **kwargs)

    async def apply_async(self, **options: Any) -> AsyncResult:
        return await self.signature(**options).apply_async()

    async def delay(self, *args: Any, **kwargs: Any) -> AsyncResult:
        return await self.apply_async(args=args, kwargs=kwargs)

    def signature(self, **options: Any) -> Signature:
        priority = first_not_null(options.get("priority"), self.priority)
        queue = first_not_null(options.get("queue"), self.queue)
        if priority is not None:
            options["priority"] = priority
        if queue is not None:
            options["queue"] = queue
        return Signature(self.name, **options)

    def s(self, *args: Any, **kwargs: Any) -> Signature:
        return self.signature(args=args, kwargs=kwargs)



from __future__ import annotations

import dataclasses
import datetime
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    import asyncio

    from .app import Celery


@dataclasses.dataclass
class RunningTask:
    asyncio_task: asyncio.Task[Any]
    task_id: str
    task_name: str
    state: Literal["SLEEPING", "SEMAPHORE", "GC", "RUNNING"]
    received: str
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    eta: str | None
    soft_time_limit: float | None
    retries: int
    started: str | None = None

    def serialize(self) -> dict[str, Any]:
        return {
            "name": self.task_name,
            "id": self.task_id,
            "state": self.state,
            "received": self.received,
            "started": self.started,
            "elapsed": (
                datetime.datetime.now().astimezone()
                - datetime.datetime.fromisoformat(self.started)
            ).total_seconds()
            if self.started is not None
            else None,
            "args": self.args,
            "kwargs": self.kwargs,
            "eta": self.eta,
            "soft_time_limit": self.soft_time_limit,
            "retries": self.retries,
            "stack": repr(self.asyncio_task.get_stack()),
        }


@dataclasses.dataclass
class _State:
    current_app: Celery | None = None
    running_tasks: dict[str, RunningTask] = dataclasses.field(default_factory=dict)
    amount_of_tasks_completed_after_last_gc_run: int = 0


_STATE = _State()


def get_current_app() -> Celery:
    app = _STATE.current_app
    if app is None:
        msg = "No current app available"
        raise RuntimeError(msg)
    return app


def set_current_app(app: Celery | None) -> None:
    _STATE.current_app = app



from __future__ import annotations

from contextvars import ContextVar

# Variables local to currently executing coroutine in the event loop
CURRENT_ROOT_ID: ContextVar[str | None] = ContextVar("current_root_id", default=None)
CURRENT_TASK_ID: ContextVar[str | None] = ContextVar("current_task_id", default=None)



from __future__ import annotations

import asyncio
import datetime
import gc
import logging
import time

from ._state import _STATE

logger = logging.getLogger(__name__)


async def perform_gc_at_regular_intervals(
    *,
    max_tasks_between_gc: int,
    max_interval_between_gc: datetime.timedelta,
    gc_is_paused: asyncio.Event,
) -> None:
    gc.disable()
    logger.warning("Automatic garbage collection disabled.")

    last_gc_time = datetime.datetime.now().astimezone()

    while True:
        await asyncio.sleep(0.5)
        if (
            _STATE.amount_of_tasks_completed_after_last_gc_run >= max_tasks_between_gc
            or (datetime.datetime.now().astimezone() - last_gc_time)
            > max_interval_between_gc
        ):
            gc_is_paused.clear()
            try:
                while True:
                    at_least_one_celery_task_is_running = any(
                        True
                        for t in _STATE.running_tasks.values()
                        if t.state == "RUNNING"
                    )
                    if not at_least_one_celery_task_is_running:
                        logger.debug("Collecting garbage manually...")
                        s = time.monotonic()
                        gc.collect()
                        logger.debug(
                            "Collecting garbage took %f seconds",
                            time.monotonic() - s,
                        )
                        last_gc_time = datetime.datetime.now().astimezone()
                        break
                    await asyncio.sleep(1)
            except Exception:
                gc.enable()
                raise
            finally:
                gc_is_paused.set()
                _STATE.amount_of_tasks_completed_after_last_gc_run = 0



from __future__ import annotations

import datetime
import json
from typing import Any

from aio_pika import DeliveryMode, Message


def create_task_message(  # noqa: PLR0913
    *,
    task_id: str,
    task_name: str,
    args: tuple[Any, ...] | None,
    kwargs: dict[str, Any] | None,
    priority: int | None,
    parent_id: str | None,
    root_id: str | None,
    chain: list[dict[str, Any]] | None = None,
    ignore_result: bool = False,
    countdown: float | None = None,
    reply_to: str = "",
    expiration: datetime.datetime | datetime.timedelta | float | None = None,
) -> Message:
    args = args or ()
    kwargs = kwargs or {}
    if countdown is None:
        eta = None
    else:
        eta = (
            datetime.datetime.now().astimezone() + datetime.timedelta(seconds=countdown)
        ).isoformat()
    # https://docs.celeryq.dev/en/latest/internals/protocol.html
    headers: dict[str, Any] = {
        "argsrepr": repr(args),
        "eta": eta,
        "expires": None,
        "group": None,  # required
        "group_index": None,
        "id": task_id,  # required
        "ignore_result": ignore_result,
        "kwargsrepr": repr(kwargs),
        "lang": "py",  # required
        "origin": "unknown_pid@unknown_hostname",
        "parent_id": parent_id,  # required
        "retries": 0,
        "root_id": root_id if root_id is not None else task_id,  # required
        "shadow": None,
        "stamped_headers": None,
        "stamps": {},
        "task": task_name,  # required
        "timelimit": [None, None],
    }
    body: bytes = json.dumps(
        (
            args,
            kwargs,
            {
                "callbacks": None,
                "errbacks": None,
                "chain": chain or None,
                "chord": None,
            },
        ),
    ).encode()
    return Message(
        body=body,
        headers=headers,
        content_type="application/json",
        content_encoding="utf-8",
        delivery_mode=DeliveryMode.PERSISTENT,
        priority=priority,
        correlation_id=task_id,
        reply_to=reply_to,  # optional
        expiration=expiration,
    )



from __future__ import annotations

import copy
import datetime
import json
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, Tuple, cast

if TYPE_CHECKING:
    from aio_pika import IncomingMessage, Message


@dataclass(frozen=True)
class Request:
    message: IncomingMessage

    id: str
    task: str
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    retries: int
    eta: datetime.datetime | None
    parent_id: str | None
    group: str | None
    root_id: str | None
    ignore_result: bool
    timelimit: tuple[int | None, int | None]
    chain: list[dict[str, Any]]

    @classmethod
    def from_message(cls: type[Request], message: IncomingMessage) -> Request:
        headers = message.headers
        args, kwargs, options = json.loads(message.body)
        raw_eta = cast(Optional[str], headers["eta"])
        eta: datetime.datetime | None
        if raw_eta is not None:
            if sys.version_info >= (3, 11):
                eta = datetime.datetime.fromisoformat(raw_eta)
            else:
                eta = datetime.datetime.strptime(raw_eta, "%Y-%m-%dT%H:%M:%S.%f%z")
            if eta.tzinfo is None or eta.tzinfo.utcoffset(eta) is None:
                msg = "eta must be a timezone aware datetime object"
                raise RuntimeError(msg)
        else:
            eta = None
        return cls(
            message=message,
            id=str(headers["id"]),
            args=args,
            kwargs=kwargs,
            task=cast(str, headers["task"]),
            retries=cast(int, headers["retries"]),
            eta=eta,
            parent_id=cast(Optional[str], headers["parent_id"]),
            group=cast(Optional[str], headers["group"]),
            root_id=cast(str, headers["root_id"]),
            ignore_result=bool(headers["ignore_result"]),
            timelimit=cast(Tuple[Optional[int], Optional[int]], headers["timelimit"]),
            chain=options["chain"],
        )

    def build_retry_message(
        self,
        *,
        countdown: datetime.timedelta | None,
    ) -> Message:
        message = copy.copy(self.message)
        if countdown is None:
            eta = None
        else:
            eta = (datetime.datetime.now().astimezone() + countdown).isoformat()
        message.headers.update(
            eta=eta,
            retries=self.retries + 1,
            parent_id=self.id,
        )
        return message



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import aio_pika
    from aio_pika.abc import (
        AbstractQueue,
        AbstractRobustChannel,
    )
    from pamqp.common import Arguments


class Broker:
    def __init__(
        self,
        *,
        broker_url: str,
        task_queue_max_priority: int | None,
        broker_publish_timeout: float | None,
        publishing_channel: AbstractRobustChannel,
    ) -> None:
        self._broker_url = broker_url
        self._task_queue_max_priority: int | None = task_queue_max_priority
        self._already_declared_queues: set[str] = set()
        self._broker_publish_timeout = broker_publish_timeout
        self._publishing_channel = publishing_channel

    async def publish_message(
        self,
        message: aio_pika.Message,
        *,
        routing_key: str,
    ) -> None:
        if routing_key not in self._already_declared_queues:
            await self.declare_queue(
                queue_name=routing_key,
                channel=self._publishing_channel,
            )

        await self._publishing_channel.default_exchange.publish(
            message,
            routing_key=routing_key,
            timeout=self._broker_publish_timeout,
        )

    async def declare_queue(
        self,
        *,
        queue_name: str,
        channel: AbstractRobustChannel,
    ) -> AbstractQueue:
        arguments: Arguments
        if self._task_queue_max_priority is not None:
            arguments = {"x-max-priority": self._task_queue_max_priority}
        else:
            arguments = None
        queue = await channel.declare_queue(
            name=queue_name,
            durable=True,
            arguments=arguments,
            timeout=self._broker_publish_timeout,
        )
        self._already_declared_queues.add(queue_name)
        return queue



from __future__ import annotations

import asyncio
import json
import time
from typing import TYPE_CHECKING, Any, Dict, cast

from .exceptions import TimeoutError as CeleryTimeoutError

if TYPE_CHECKING:
    from .app import Celery


class AsyncResult:
    def __init__(
        self,
        id: str,  # noqa: A002
        *,
        app: Celery,
    ) -> None:
        self.id = id
        self._cache = None
        self.result_backend = app.result_backend

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.id}>"

    async def _get_task_meta(self) -> dict[str, Any]:
        if self.result_backend is None:
            msg = "No result backend is configured."
            raise RuntimeError(msg)
        if self._cache is None:
            value = await self.result_backend.get(f"celery-task-meta-{self.id}")
            if value is None:
                return {"result": None, "status": "PENDING"}
            self._cache = json.loads(value)
        return cast(Dict[str, Any], self._cache)

    @property
    async def result(self) -> Any:
        return (await self._get_task_meta())["result"]

    @property
    async def state(self) -> str:
        return str((await self._get_task_meta())["status"])

    async def get(
        self,
        timeout: float | None = None,  # noqa: ASYNC109
        interval: float = 0.5,
    ) -> Any:
        """Wait until task is ready, and return its result."""
        value = await self._get_task_meta()
        start = time.monotonic()
        while value == {"result": None, "status": "PENDING"}:
            await asyncio.sleep(interval)
            if timeout is not None and (time.monotonic() - start) > timeout:
                msg = "The operation timed out."
                raise CeleryTimeoutError(msg)
            value = await self._get_task_meta()
        return value["result"]
